package SPVM::Sys::Resource::Windows;



1;

=head1 Name

SPVM::Sys::Resource::Windows - Resource for Windows

=head1 Description

Sys::Resource::Windows class in L<SPVM> is a resource for Windows.

=head1 Usage

  use Sys::Resource::Windows;

=head1 See Also

=over 2

=item * L<Sys|SPVM::Sys>

=back

=head1 Copyright & License

Copyright (c) 2025 Yuki Kimoto

MIT License

